"""
This module performs all basic SFA operations.
It is an interface for sfa automata.
"""
# !/usr/bin/python
import random
import string

from dfa import DFA


class Predicate:
    """
    The predicate statement
    """

    def __init__(self, initializer):
        """
        This __init__ method is not implemented
        """
        raise NotImplementedError('__init__ method not implemented')

    def is_sat(self, symbol):
        """
        This is_sat method is not implemented
        """
        raise NotImplementedError('is_sat method not implemented')

    def refactor(self, symbol, value):
        """
        This refactor method is not implemented
        """
        raise NotImplementedError('refactor method not implemented')

    def get_witness(self):
        """
        This get_witness method is not implemented
        """
        raise NotImplementedError('get_witness method not implemented')

    def __iter__(self):
        """
        This __iter__ method is not implemented
        """
        raise NotImplementedError('__iter__ method not implemented')


class SetPredicate(Predicate):
    """
    This class initializes and sets a predicate for the
    current transition
    """

    def __init__(self, initializer):
        """
        Initialization Function
        Args:
            initializer (list): A list of characters
        Returns:
            None
        """
        self.pset = set(initializer)

    def is_sat(self, symbol):
        """
        Args:
            symbol:
        Returns:
            bool: The response can be true or false
        """
        return symbol in self.pset

    def refactor(self, symbol, value):
        """
        Args:
            symbol:
            value:
        Returns:
            None
        """
        if value:
            self.pset.add(symbol)
        else:
            self.pset.remove(symbol)

    def get_witness(self):
        """
        Args:
            None
        Returns:
            str: A random sample
        """
        return random.sample(self.pset, 1)[0]

    def __iter__(self):
        """
        Args:
            None
        Returns:
            None
        """
        return iter(self.pset)


class SFAState:
    """The SFA state structure"""
    def __init__(self, sid=None):
        """
        Args:
            sid (int): The state identifier
        Returns:
            None
        """
        self.final = False
        self.initial = False
        self.state_id = sid
        self.arcs = []

    def __iter__(self):
        """
        Args:
            None
        Returns:
            None
        """
        return iter(self.arcs)


class SFAArc:
    """The SFA Arc structure"""
    def __init__(self, src_state_id, dst_state_id, guard_p, term=None):
        """
        Initialization function for Arc's guardgen structure
        Args:
            src_state_id (int): The source state identifier
            dst_state_id (int): The destination state identifier
            guard_p: The input character
            term: The input term
        Returns:
            None
        """
        self.src_state = src_state_id
        self.dst_state = dst_state_id
        self.guard = guard_p
        self.term = None


class SFA:
    """
       Symbolic Finite Automata (SFAs) are finite state symautomata
       in which the alphabet is given by a Boolean algebra that may
       have an infinite domain, and transitions are labeled with
       first-order predicates over such algebra
    """

    def __init__(self, alphabet=None):
        """
        Initialization of the SFA oject
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        self.states = []
        self.arcs = []
        self.alphabet = alphabet

    def add_state(self):
        """This function adds a new state"""
        sid = len(self.states)
        self.states.append(SFAState(sid))

    def add_arc(self, src, dst, char):
        """
        This function adds a new arc in a SFA state
        Args:
            src (int): The source state identifier
            dst (int): The destination state identifier
            char (str): The transition symbol
        Returns:
            None
        """
        assert type(src) == type(int()) and type(dst) == type(int()), \
            "State type should be integer."
        while src >= len(self.states) or dst >= len(self.states):
            self.add_state()
        self.states[src].arcs.append(SFAArc(src, dst, char))

    def random_string(self, size=1):
        """
        This function generates a random string
        Args:
            size (int): The length of the random string
        Returns:
            str: A random string
        """
        pass

    def consume_input(self, inp):
        """
        Return True/False if the machine accepts/reject the input.
        Args:
            inp (str): input string to be consumed
        Retunrs:
            bool: A true or false value depending on if the DFA
                accepts the provided input
        """
        cur_state = self.states[0]
        for character in inp:
            found = False
            for arc in cur_state.arcs:
                if arc.guard.is_sat(character):
                    cur_state = self.states[arc.dst_state]
                    found = True
                    break

            if not found:
                raise RuntimeError('SFA not complete')

        return cur_state.final

    def concretize(self):
        """
        Transforms the SFA into a DFA
        Args:
            None
        Returns:
            DFA: The generated DFA
        """
        dfa = DFA(self.alphabet)
        for state in self.states:
            for arc in state.arcs:
                for char in arc.guard:
                    dfa.add_arc(arc.src_state, arc.dst_state, char)

        for i in xrange(len(self.states)):
            if self.states[i].final:
                dfa[i].final = True
        return dfa


    def save(self, txt_fst_filename):
        """
        Save the machine in the openFST format in the file denoted by
        txt_fst_filename.
        Args:
            txt_fst_filename (str): The name of the file
        Returns:
            None
        """
        dfa = self.concretize()
        return dfa.save(txt_fst_filename)


    def load(self, txt_fst_filename):
        """
        Save the transducer in the text file format of OpenFST.
        The format is specified as follows:
            arc format: src dest ilabel olabel [weight]
            final state format: state [weight]
        lines may occur in any order except initial state must be first line
        Args:
            txt_fst_filename (string): The name of the file
        Returns:
            None
        """
        raise NotImplementedError('SFA load method not implemented')




def main():
    """Main Function"""
    alphabet = list(string.lowercase)  # + ["<", ">"]

    # Create an SFA for the regular expression .*<t>.*
    sfa = SFA(alphabet)

    # sfa.add_arc(0,0,SetPredicate([ i for i in alphabet if i != "<" ]))
    # sfa.add_arc(0,1,SetPredicate(list("<")))
    #
    # sfa.add_arc(1,2,SetPredicate(list("t")))
    # sfa.add_arc(1,0,SetPredicate([ i for i in alphabet if i != "t" ]))
    #
    # sfa.add_arc(2,3,SetPredicate(list(">")))
    # sfa.add_arc(2,0,SetPredicate([ i for i in alphabet if i != ">" ]))
    #
    # sfa.add_arc(3,3,SetPredicate(alphabet))
    #
    # sfa.states[3].final = True

    sfa.add_arc(0, 7, SetPredicate([i for i in alphabet if i != "d" and i != "input_string"]))
    sfa.add_arc(1, 7, SetPredicate([i for i in alphabet if i != "i"]))
    sfa.add_arc(2, 7, SetPredicate([i for i in alphabet if i != "p"]))
    sfa.add_arc(3, 7, SetPredicate([i for i in alphabet if i != "v"]))
    sfa.add_arc(5, 7, SetPredicate(list(alphabet)))
    sfa.add_arc(4, 7, SetPredicate([i for i in alphabet if i != "a"]))
    sfa.add_arc(6, 7, SetPredicate([i for i in alphabet if i != "n"]))
    sfa.add_arc(7, 7, SetPredicate(list(alphabet)))

    sfa.add_arc(0, 1, SetPredicate(list("d")))
    sfa.add_arc(1, 3, SetPredicate(list("i")))
    sfa.add_arc(3, 5, SetPredicate(list("v")))

    sfa.add_arc(0, 2, SetPredicate(list("input_string")))
    sfa.add_arc(2, 4, SetPredicate(list("p")))
    sfa.add_arc(4, 6, SetPredicate(list("a")))
    sfa.add_arc(6, 5, SetPredicate(list("n")))

    sfa.states[5].final = True

    dfa = sfa.concretize()
    # dfa.minimize()
    dfa.save('concrete_re_sfa.dfa')

    # Consume some input
    input_string = "koukouroukou"
    print 'SFA-DFA result on {}: {} - {}'.format(input_string, sfa.consume_input(input_string),
                                                 dfa.consume_input(input_string))

    input_string = "divspan"
    print 'SFA-DFA result on {}: {} - {}'.format(input_string, sfa.consume_input(input_string),
                                                 dfa.consume_input(input_string))

    input_string = "div"
    print 'SFA-DFA result on {}: {} - {}'.format(input_string, sfa.consume_input(input_string),
                                                 dfa.consume_input(input_string))

    input_string = "span"
    print 'SFA-DFA result on {}: {} - {}'.format(input_string, sfa.consume_input(input_string),
                                                 dfa.consume_input(input_string))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import ez_setup
import sys
import os
import imp
import tempfile
ez_setup.use_setuptools()

PROJECT = 'symautomata'

# Change docs/sphinx/conf.py too!
VERSION = '0.0.4'

from setuptools import setup, find_packages

try:
    long_description = open('README.md', 'rt').read()
except IOError:
    long_description = ''

setup(
    name=PROJECT,
    version=VERSION,

    description='Automata classes',
    long_description=long_description,

    author='George Argyros, Ioannis Stais',
    author_email='ioannis.stais@gmail.com',

    url='https://github.com/GeorgeArgyros/symautomata',
    download_url='https://github.com/GeorgeArgyros/symautomata/tarball/master',

    classifiers=['Development Status :: 3 - Alpha',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 2',
                 'Programming Language :: Python :: 2.7',
                 'Intended Audience :: Developers',
                 'Environment :: Console',
                 ],

    platforms=['Any'],
    scripts=[],
    provides=[],
    install_requires=['FAdo', 'python-dateutil'],
    namespace_packages=[],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
)


installpywrapfst1 = """
ruby -e "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/master/install)"
brew doctor

sudo easy_install pip

brew install wget

brew install graphviz

cd ~
mkdir -p prefix
cd prefix
export PREFIX=$PWD
echo $PREFIX

mkdir -p openfst
cd openfst
wget http://www.openfst.org/twiki/pub/FST/FstDownload/openfst-1.5.4.tar.gz
tar xzf openfst-1.5.4.tar.gz

rm -rf objdir
mkdir objdir
cd objdir
../openfst-1.5.4/configure --prefix=$PREFIX --enable-pdt --enable-bin --enable-ngram-fsts  --enable-python
make -j 4
make install

echo "export PREFIX=$(cd; pwd)/prefix" >> ~/.bashrc

echo 'export CPLUS_INCLUDE_PATH=$PREFIX/include:$CPLUS_INCLUDE_PATH
export LIBRARY_PATH=$PREFIX/lib:$LIBRARY_PATH
export LIBRARY_PATH=$PREFIX/lib/fst:$LIBRARY_PATH
export LD_LIBRARY_PATH=$PREFIX/lib:$LD_LIBRARY_PATH
export LD_LIBRARY_PATH=$PREFIX/lib/fst:$LD_LIBRARY_PATH
export DYLD_LIBRARY_PATH=$PREFIX/lib/fst:$PREFIX/lib:$DYLD_LIBRARY_PATH
export PATH=$PREFIX/bin:$PATH
export PYTHONPATH=$PREFIX/lib/python2.7/site-packages:$PYTHONPATH' >> ~/.bashrc

echo "source ~/.bashrc" >> ~/.bash_profile

echo $PREFIX

for f in `find $PREFIX/include/fst -type f | xargs grep isspace | cut -f 1 -d ':' | uniq`; do
  sed -i .bak 's/isspace/std::isspace/g' $f
done
pip install openfst
"""
installpywrapfst2="""
wget http://www.openfst.org/twiki/pub/FST/FstDownload/openfst-1.5.4.tar.gz
tar zxvf openfst-1.5.4.tar.gz
cd openfst-1.5.4
./configure --enable-python
make
sudo make install
"""


def check_for_fst():
    try:
        print 'Checking for fst module:',
        imp.find_module('fst')
        print 'OK'
    except ImportError:
        print 'FAIL'
        try:
            print 'Checking for pywrapfst module:',
            imp.find_module('pywrapfst')
            print 'OK'
        except ImportError:
            print 'FAIL'
            print 'It is recommended to use openfst python bindings (either pyfst or pywrapfst)' \
                  ' for the DFA implementation. While this is not necessary, using openfst python' \
                  ' bindings will increase execution speed significally.'
            install = raw_input(
                ('* Install pywrapfst now? [y/n] ')
            )
            if install == 'y':
                os.system(installpywrapfst2)
                try:
                    print 'Checking for pywrapfst module:',
                    imp.find_module('pywrapfst')
                    print 'OK'
                except ImportError:
                    print 'FAIL'
                    os.system(installpywrapfst1)
                    try:
                        print 'Checking again for pywrapfst module:',
                        imp.find_module('pywrapfst')
                    except ImportError:
                        print 'FAIL'
                        print 'The application will fallback to a python implementation during execution'
            else:
                print 'The application will fallback to a python implementation during execution'

check_for_fst()



"""This modules generates a string from a CFG"""


class CFGGenerator(object):
    """This class generates a string from a CFG"""
    grammar = None
    resolved = None
    bfs_queue = None
    maxstate = 0

    def __init__(self, cfgr=None, optimized=1, splitstring=0, maxstate=0):
        """
        Object initialization
        Args:
            cfgr (CNF): grammar for the random objects
            optimized (bool): mode of operation - if enabled not all
                            CNF rules are included (mitigate O(n^3))
            splitstring (bool): A boolean for enabling or disabling
                            the splitting of symbols using a space
            maxstate (int): The maxstate is used for generating in a
                            dynamic way the CNF rules that were not
                            included due to the optimization. As a
                            result, the algorithm generates these
                            rules only if required.
        Returns:
            None
        """
        self.grammar = None
        self.resolved = None
        self.bfs_queue = None
        self.maxstate = 0
        self.grammar = cfgr
        self.optimized = optimized
        self.splitstring = splitstring
        self.grammar.terminalrules()
        # Because of the optimization, there are some non
        # existing terminals on the generated list
        if self.optimized:
            self._clean_terminals()
        self.grammar.nonterminalrules()
        self.resolved = {}
        self.bfs_queue = []
        self.maxstate = maxstate

    def generate(self):
        """
        Generates a new random string from the start symbol
        Args:
            None
        Returns:
            str: The generated string
        """

        result = self._gen(self.optimized, self.splitstring)
        if self.splitstring and result is not None:
            result = result[1:]
        return result

    def _clean_terminals(self):
        """
        Because of the optimization, there are some non existing terminals
        on the generated list. Remove them by checking for terms in form Ax,x
        """
        new_terminals = []
        for term in self.grammar.Terminals:
            x_term = term.rfind('@')
            y_term = term.rfind('A')
            if y_term > x_term:
                x_term = y_term
            ids = term[x_term + 1:].split(',')
            if len(ids) < 2:
                """It'input_string a normal terminal, not a state"""
                new_terminals.append(term)
        self.grammar.Terminals = new_terminals

    def _check_self_to_empty(self, stateid):
        """
        Because of the optimization, the rule for empty states is missing
        A check takes place live
        Args:
            stateid (int): The state identifier
        Returns:
            bool: A true or false response
        """
        x_term = stateid.rfind('@')
        y_term = stateid.rfind('A')
        if y_term > x_term:
            x_term = y_term
        ids = stateid[x_term + 1:].split(',')
        if len(ids) < 2:
            return 0
        if ids[0] == ids[1]:
            #    print 'empty'
            return 1
        return 0

    def _check_intemediate(self, myntr, maxstate):
        """
        For each state Apq which is a known terminal, this function
        searches for rules Apr -> Apq Aqr and Arq -> Arp Apq where
        Aqr is also a known terminal or Arp is also a known terminal.
        It is mainly used as an optimization in order to avoid the O(n^3)
        for generating all the Apq -> Apr Arq rules during the PDA to CFG
        procedure.
        Args:
            myntr (str): The examined non terminal that was poped out
                        of the queue
            maxstate (int): The maxstate is used for generating in a
                            dynamic way the CNF rules that were not
                            included due to the optimization. As a
                            result, the algorithm generates these
                            rules only if required.
        Returns:
            bool: Returns true if the algorithm was applied
                at least one time
        """
        # print 'BFS Dictionary Update - Intermediate'
        x_term = myntr.rfind('@')
        y_term = myntr.rfind('A')
        if y_term > x_term:
            x_term = y_term
        ids = myntr[x_term + 1:].split(',')
        if len(ids) < 2:
            return 0
        i = ids[0]
        j = ids[1]
        r = 0
        find = 0
        while r < maxstate:
            if r != i and r != j:
                if 'A' + i + ',' + \
                        repr(r) not in self.resolved \
                        and 'A' + j + ',' + repr(r) in self.resolved:
                    self.resolved[
                        'A' + i + ',' + repr(r)] = self.resolved[myntr] \
                                                   + self.resolved['A' + j + ',' + repr(r)]
                    if self._checkfinal('A' + i + ',' + repr(r)):
                        return self.resolved['A' + i + ',' + repr(r)]
                    if 'A' + i + ',' + repr(r) not in self.bfs_queue:
                        self.bfs_queue.append('A' + i + ',' + repr(r))
                    find = 1
                if 'A' + repr(r) + ',' + j not in self.resolved and 'A' + \
                        repr(r) + ',' + i in self.resolved:
                    self.resolved[
                        'A' + repr(r) + ',' + j] = self.resolved['A' + repr(r) + ',' + i] \
                                                   + self.resolved[myntr]
                    if self._checkfinal('A' + repr(r) + ',' + j):
                        return self.resolved['A' + repr(r) + ',' + j]
                    if 'A' + repr(r) + ',' + j not in self.bfs_queue:
                        self.bfs_queue.append('A' + repr(r) + ',' + j)
                    find = 1
            r = r + 1
        if find == 1:
            return 1
        return 0

    def _check_self_replicate(self, myntr):
        """
        For each Rule B -> c where c is a known terminal, this function
        searches for B occurences in rules with the form A -> B and sets
        A -> c.
        """
        # print 'BFS Dictionary Update - Self Replicate'
        find = 0
        for nonterm in self.grammar.ntr:
            for i in self.grammar.ntr[nonterm]:
                if self.grammar.Rules[i][0] not in self.resolved and not isinstance(
                        self.grammar.Rules[i][1], (set, tuple)) \
                        and self.grammar.Rules[i][1] == myntr:
                    self.resolved[self.grammar.Rules[i][0]] = self.resolved[myntr]
                    if self._checkfinal(self.grammar.Rules[i][0]):
                        return self.resolved[self.grammar.Rules[i][0]]
                    if self.grammar.Rules[i][0] not in self.bfs_queue:
                        self.bfs_queue.append(self.grammar.Rules[i][0])
                    find = 1
        if find == 1:
            return 1
        return 0

    def _check_self_nonterminals(self, optimized):
        """
        For each Rule A -> BC where B and C are known terminals (B -> c1 and C -> c2),
        this function searches replaces A to the corresponding terminals A -> c1c2
        """
        # print 'BFS Dictionary Update - Self Non Terminals'
        find = 0
        for nt in self.grammar.ntr:
            for i in self.grammar.ntr[nt]:
                if (self.grammar.Rules[i][0] not in self.resolved\
                    or self.grammar.Rules[i][0] == 'S') \
                    and isinstance(self.grammar.Rules[i][1], (set, tuple)):
                    # All rules are in CNF form, so first check the A -> BC rules
                    part_a = None
                    if optimized and self._check_self_to_empty(
                            self.grammar.Rules[i][1][0]):
                        part_a = ''
                    elif self.grammar.Rules[i][1][0] in self.resolved:
                        part_a = self.resolved[self.grammar.Rules[i][1][0]]
                    part_b = None
                    if optimized and self._check_self_to_empty(
                            self.grammar.Rules[i][1][1]):
                        part_b = ''
                    elif self.grammar.Rules[i][1][1] in self.resolved:
                        part_b = self.resolved[self.grammar.Rules[i][1][1]]
                    if part_a is not None and part_b is not None:
                        self.resolved[self.grammar.Rules[i][0]] = part_a + part_b
                        # print 'Non Terminals Resolving
                        # '+self.g.Rules[i][0]+": "+
                        # self.Resolved[self.g.Rules[i][0]]
                        if self._checkfinal(self.grammar.Rules[i][0]):
                            return self.resolved[self.grammar.Rules[i][0]]
                        if self.grammar.Rules[i][0] not in self.bfs_queue:
                            self.bfs_queue.append(self.grammar.Rules[i][0])
                        find = 1

        if find == 1:
            return 1
        return 0

    def _checkfinal(self, nonterminal):
        if nonterminal == 'S':
            return 1
        return 0

    def _gen(self, optimized, splitstring):
        """Generates a new random object generated from the nonterminal
        Args:
            optimized (bool): mode of operation - if enabled not all
                            CNF rules are included (mitigate O(n^3))
            splitstring (bool): A boolean for enabling or disabling
        Returns:
            str: The generated string
        """
        # Define Dictionary that holds resolved rules
        # (only in form A -> terminals sequence)
        self.resolved = {}
        # First update Resolved dictionary by adding rules
        # that contain only terminals (resolved rules)
        for nt in self.grammar.ntr:
            for i in self.grammar.ntr[nt]:
                if self.grammar.Rules[i][0] not in self.resolved\
                        and not isinstance(self.grammar.Rules[i][1], (set, tuple)):
                    if self.grammar.Rules[i][1] != '@empty_set' \
                            and self.grammar.Rules[i][1] in self.grammar.Terminals:

                        if splitstring:
                            self.resolved[
                                self.grammar.Rules[i][0]] = self.grammar.Rules[i][1]
                        else:
                            if self.grammar.Rules[i][1] == '&':
                                self.resolved[self.grammar.Rules[i][0]] = ' '
                            else:
                                self.resolved[
                                    self.grammar.Rules[i][0]] = self.grammar.Rules[i][1]
                        # print 'ResolvingA '+self.g.Rules[i][0]+": "+
                        # self.g.Rules[i][1]
                        if self._checkfinal(self.grammar.Rules[i][0]):
                            return self.resolved[self.grammar.Rules[i][0]]
                        if self.grammar.Rules[i][0] not in self.bfs_queue:
                            self.bfs_queue.append(self.grammar.Rules[i][0])
                    if self.grammar.Rules[i][1] == '@empty_set':
                        self.resolved[self.grammar.Rules[i][0]] = ''
                        # print 'ResolvingB '+self.g.Rules[i][0]+": "
                        self.bfs_queue.append(self.grammar.Rules[i][0])
                    if optimized and self._check_self_to_empty(
                            self.grammar.Rules[i][1]):
                        self.resolved[self.grammar.Rules[i][0]] = ''
                        # print 'ResolvingC '+self.g.Rules[i][0]+": "
                        if self.grammar.Rules[i][0] not in self.bfs_queue:
                            self.bfs_queue.append(self.grammar.Rules[i][0])

        # Then try to use the rules from Resolved dictionary and check
        # if there is another rule that can be resolved.
        # This should be done in a while loop
        change = 1
        while change:
            change = 0
            if not change:
                ret = self._check_self_nonterminals(optimized)
                if ret == 1:
                    change = 1
                elif ret != 0:
                    return ret
            if not change:
                while not change and len(self.bfs_queue) > 0:
                    myntr = self.bfs_queue.pop()
                    ret = self._check_self_replicate(myntr)
                    if ret == 1:
                        change = 1
                    elif ret != 0:
                        return ret
                    if optimized and self._check_intemediate(
                            myntr, self.maxstate):
                        change = 1
                        break
